# frozen_string_literal: true

require [Rails.root, %W[lib core_extentions parameters querestrict]].flatten.join('/')

module Hermes
  module Tools
    module Aggregations
      SCORING_PARAMS = %i[
        profile_epci_score
        eligible_position
        applicated_position
        effort_rate
      ]

      def format_params(fields, param)
        params.each_with_object({}) do |(key, val), h|
          next unless fields[key]

          h[key.to_sym] = format_parameter(fields[key].type.to_s, val)
        end
      end

      def aggregate_with_score(target_class_short, main_id, params, base_aggregation = {})
        target_class, foreign_key, option_query, main_field = case target_class_short
                                                              when 'requests'
                                                                [
                                                                  Hermes::HousingRequest,
                                                                  'main_applicant',
                                                                  {
                                                                    is_active: true,
                                                                    :$expr => {
                                                                      :$eq => [
                                                                        '$main_applicant_id',
                                                                        '$$foreign_id'
                                                                      ]
                                                                    }
                                                                  },
                                                                  :housing_offer_id
                                                                ]
                                                              when 'offers'
                                                                [
                                                                  Hermes::HousingOffer,
                                                                  'housing_offer',
                                                                  {
                                                                    :$expr => {
                                                                      :$eq => [
                                                                        '$_id',
                                                                        '$$foreign_id'
                                                                      ]
                                                                    }
                                                                  },
                                                                  :applicant_id
                                                                ]
                                                              when 'applications'
                                                                [
                                                                  Hermes::HousingOffer,
                                                                  'housing_offer',
                                                                  {
                                                                    :$expr => {
                                                                      :$eq => [
                                                                        '$_id',
                                                                        '$$foreign_id'
                                                                      ]
                                                                    }
                                                                  },
                                                                  :housing_offer_id
                                                                ]
                                                              end
        target_fields = target_class.fields
        sorting_params = params[:sort]
        query_params = params.reject { |key| key != :sort && key != 'sort' }
        scoring_params = format_params(Hermes::Score.fields, query_params.slice(*SCORING_PARAMS))
        target_params = format_params(
          target_fields,
          query_params.reject { |key| SCORING_PARAMS.include? key }
        )

        scores_query = {
          main_field => main_id,
          is_active: true
        }.merge(scoring_params)

        pipeline = [{ :$match => scores_query }]
        foreign_pipeline = [{ :$match => option_query.merge(base_aggregation.selector) }]
        pipeline << {
          :$lookup => {
            from: target_class.collection.namespace.sub(/document(_test)?\./, ''),
            let: { foreign_id: "$#{foreign_key.sub('main_', '')}_id" },
            pipeline: foreign_pipeline,
            as: 'foreignItems'
          }
        }
        pipeline << {
          :$addFields => { foreignItem: { :$arrayElemAt => ['$foreignItems', 0] } }
        }
        if sorting_params.present?
          formated_sorting = sorting_params.each_with_object({}) do |(key, val), h|
            key = "foreignItem.#{key}".to_sym unless SCORING_PARAMS.include?(key.to_sym)
            h[key] = val
          end
          # to reject elements without matching lookup :
          pipeline << { :$match => { foreignItems: { :$exists => true, :$size => 1 } } }
          pipeline << {
            :$sort => handle_sort(formated_sorting)
          }
        end

        page = params[:page]&.to_i || 1
        perPage = params[:per_page]&.to_i || 15
        skipPage = (page - 1) * perPage
        pipeline << {
          :$facet  => {
            :paginatedResults =>[{ :$skip  => skipPage }, { :$limit  => perPage }],
            :totalCount => [{ :$count => 'count' } ]
          }
        }

        Hermes::Score.collection.aggregate(pipeline)
      end
    end
  end
end

include Hermes::Tools::Aggregations
